<?php
/**
 * @package     HikaShop for Joomla!
 * @version     2.4.1
 * @author		hikashop.com
 * @copyright   (C) 2010-2018 HIKARI SOFTWARE. All rights reserved.
 * @license     GNU/GPLv3 http://www.gnu.org/licenses/gpl-3.0.html
 *
 * Modified for QuickPay payment gateway by:
 * info@stokersoft.com		20/11-2014		-
 * rasmus@earmaster.com 	05/12-2015		-
 * mko@f-engel.com			24/05-2018		Added API integration for "Capture" and "Refund" through HikaShop.
 */
defined('_JEXEC') or die('Restricted access');
use Joomla\CMS\Factory;

class plgHikashoppaymentQuickpay extends hikashopPaymentPlugin
{
    var $accepted_currencies = array(
        'AUD','BRL','CAD','EUR','GBP','JPY','USD','NZD','CHF','HKD','SGD','SEK',
        'DKK','PLN','NOK','HUF','CZK','MXN','MYR','PHP','TWD','THB','ILS','TRY'
    );

    var $multiple    = true;
    var $name        = 'quickpay';
    var $doc_form    = 'quickpay';
    var $ordernumber = '';

    function __construct(&$subject, $config) {
        parent::__construct($subject, $config);
    }

    function onBeforeOrderCreate(&$order,&$do){
        if(parent::onBeforeOrderCreate($order, $do) === true)
            return true;
    }

    function onAfterOrderConfirm(&$order, &$methods, $method_id) {
        parent::onAfterOrderConfirm($order, $methods, $method_id);
        $hkconfig =& hikashop_config();

        if($this->currency->currency_locale['int_frac_digits'] > 2) {
            $this->currency->currency_locale['int_frac_digits'] = 2;
        }

        $notify_url = HIKASHOP_LIVE.'index.php?option=com_hikashop&ctrl=checkout&task=notify&notif_payment='.$this->name.'&tmpl=component&lang='.$this->locale . $this->url_itemid;
        $return_url = HIKASHOP_LIVE.'index.php?option=com_hikashop&ctrl=checkout&task=after_end&order_id='.$order->order_id . $this->url_itemid;
        $cancel_url = HIKASHOP_LIVE.'index.php?option=com_hikashop&ctrl=order&task=cancel_order&order_id='.$order->order_id . $this->url_itemid;

        $tax_total      = '';
        $discount_total = '';
        $debug          = @$this->payment_params->debug;

        if(!isset($this->payment_params->no_shipping))
            $this->payment_params->no_shipping = 1;
        if(!empty($this->payment_params->rm))
            $this->payment_params->rm = 2;

        $amount      = (int)($order->cart->full_total->prices[0]->price_value_with_tax * 100);
        $ordernumber = $order->order_number;
        $orderid     = $order->order_id;

        if($this->payment_params->language == 'auto') {
            // Languages QuickPay is available in
            $availableLanguages = array("da", "de", "en", "es", "fo", "fi", "fr", "kl", "it", "no", "nl", "pl", "ru", "sv");

            // Fallback language set in params
            $fallbackLanguage = $this->payment_params->fallback_language;

            // Fetches language Joomla is set to display in.
            $langContainer = Factory::getLanguage();

            // Takes the language out of the string
            $JLang = substr($langContainer->getTag(), 0, 2);

            // Checks whether QuickPay has been translated to detected language
            if(in_array($JLang, $availableLanguages)){
                $language = $JLang;
            }else{
                // Falls back to default language if translation is not available
                $language = isset($fallbackLanguage) ? $fallbackLanguage : 'en';
            }
        } else {
            // Use set language
            $language = $this->payment_params->language;
        }

        $api_key = $this->payment_params->api_key;

        // Branding / Theme
        $branding = $this->payment_params->branding_id;

        // Prepares the form data.
        $params = array
        (
            'version' 			  => 'v10',
            'merchant_id' 		  => $this->payment_params->merchant,
            'agreement_id'	 	  => $this->payment_params->agreement_id,
            'order_id' 			  => $this->payment_params->orderid_prefix . $orderid,
            'amount' 			  => $amount,
            'currency' 			  => $this->currency->currency_code,
            'callbackurl' 		  => $notify_url,
            'continueurl' 		  => $return_url,
            'cancelurl' 		  => $cancel_url,
            'language' 			  => $language,
            'autocapture' 		  => $this->payment_params->autocapture,
            'payment_methods' 	  => $this->payment_params->cardtypelock,
            'branding_id'		  => $branding,
            'shopsystem[name]'    => 'HikaShop',
            'shopsystem[version]' => $hkconfig->get('version')
        );

        if ($this->payment_params->resurs_bank == 1) {
            $basket = array();
            for ($i = 0; $i < count($order->cart->products); $i ++) {
                $basket[$i]['qty'] = $order->cart->products[$i]->order_product_quantity;
                $basket[$i]['item_no'] = $order->cart->products[$i]->order_product_code;
                $basket[$i]['item_name'] = $order->cart->products[$i]->order_product_name;
                $basket[$i]['item_price'] = (int) ($order->cart->products[$i]->order_product_price * 100);
                $basket[$i]['vat_rate'] = (int)number_format($order->cart->products[$i]->order_product_tax * 100, 2, '.', ',');
            }

            $invoice = array();
            $invoice['name']    = $order->cart->usable_addresses->billing[1]->address_lastname;
            $invoice['street']  = $order->cart->usable_addresses->billing[1]->address_street;
            $invoice['city']    = $order->cart->usable_addresses->billing[1]->address_city;
            $invoice['country_code'] = $order->cart->billing_address->address_country->zone_code_3;
            $invoice['zip_code']    = $order->cart->usable_addresses->billing[1]->address_post_code;
            $invoice['email']       = $order->customer->user_email;
            $invoice['phone_number'] = $order->cart->usable_addresses->billing[1]->address_telephone;

            $requireFields = array(
                'basket' => $basket,
                'invoice_address' => $invoice
            );

            $params = array_merge($params, $requireFields);
        }

        // Custom Variables
        $params['variables[CUSTOM_orderid]'] = $orderid;

        // Generate Checksum
        $params["checksum"] = $this->sign_send($params, $api_key);

        if(empty($this->payment_params->url)) {
            $this->payment_params->url = 'https://payment.quickpay.net';
        }

        $this->payment_params->cardtypelock = 'creditcard';

        $this->vars = $params;

        return $this->showPage('end');
    }

    function onPaymentNotification(&$statuses) {
        // Fetches & decodes the callback data
        $raw_callback_data = file_get_contents("php://input");
        $callback_data     = json_decode($raw_callback_data);

        // Loads the order data from HikaShop
        $order_id = @$callback_data->variables->CUSTOM_orderid; // why not just used order_id ..?
        $dbOrder  = $this->getOrder($order_id);

        $this->loadPaymentParams($dbOrder);
        $this->loadOrderData($dbOrder);

        // Prepares & calculates the checksum
        $private_key = $this->payment_params->private_key;
        $checksum    = $this->sign_received($raw_callback_data, $private_key);

        // Set mail to user's language - variables are used elsewhere by HikaShop
        $dbOrder->customer = hikashop_get('class.user')->get($dbOrder->order_user_id);
        $locale            = hikashop_get('class.order')->loadLocale($dbOrder);

        // Preps & Configures the mailer
        $mailer = Factory::getMailer();
        $config = Factory::getConfig();

        $sender = array(
            $config->get('mailfrom'),
            $config->get('fromname')
        );

        // Check whether test mode is enabled.
        $test_mode = $this->payment_params->testmode;
        $test_data = isset($callback_data->test_mode) ? $callback_data->test_mode : FALSE;

        if($test_data === FALSE) {
            $proceed = TRUE;
        } else {
            if($test_mode == 1) {
                $proceed = TRUE;
            } else {
                $proceed = FALSE;
            }
        }

        // Checks whether user wants to receive alerts
        $alert_on_error = $this->payment_params->receive_alert;

        // If user has written a email address to send to, use it, otherwise send to Joomla admin (Only if alert_on_error is turned on)
        if(empty($this->payment_params->alert_email)) {
            $recipient = $config->get( 'mailfrom' );
        } else {
            if (!filter_var($this->payment_params->alert_email, FILTER_VALIDATE_EMAIL)) {
                $alert_on_error = 0;
            } else {
                $recipient = $this->payment_params->alert_email;
            }
        }

        // Preps mail data for alert
        $mailer->setSender($sender);
        $mailer->addRecipient($recipient);
        $mailer->isHTML(true);
        $mailer->Encoding = 'base64';

        $mail_body = '<h2>Alert received from: ' . HIKASHOP_LIVE . ' - (' . date('H:i - d/m/Y') . ')</h2>'
            . '<p>'
            . '<b>User IP</b>: ' . $callback_data->metadata->customer_ip . '<br />'
            . '<b>Order ID</b>: %orderid% <br />'
            . '<b>Order number</b>: %ordernumber% <br />'
            . '<b>Error message</b>:<br /> %errormsg% '
            . '</p>';

        $status_codes = array
        (
            '20000' => 'Approved',
            '40000' => 'Rejected By Acquirer',
            '40001' => 'Request Data Error',
            '50000' => 'Gateway Error',
            '50300' => 'Communications Error (with Acquirer)'
        );

        // Fetches the operation data from the callback
        $operationData = $callback_data->operations;
        $operRunID     = (count($operationData)) - 1; //Array starts from 0

        // Prepares a history note for the order
        $email   = new stdClass();
        $history = new stdClass();
        $history->amount = @$callback_data->balance . @$callback_data->currency;

        // Prepares mail data for the admin notification
        $url         = HIKASHOP_LIVE.'administrator/index.php?option=com_hikashop&ctrl=order&task=edit&order_id='.$order_id;
        $order_text  = "\r\n".JText::sprintf('NOTIFICATION_OF_ORDER_ON_WEBSITE',$dbOrder->order_number,HIKASHOP_LIVE);
        $order_text .= "\r\n".str_replace('<br/>',"\r\n",JText::sprintf('ACCESS_ORDER_WITH_LINK',$url));

        // Checks that the received order number / ID matches the one stored in the database, if alert is turned on, send alert.
        if($dbOrder->order_id != $callback_data->variables->CUSTOM_orderid) {
            if($alert_on_error == 1) {
                $error_msg = 'The order number received did not match with the ID, the incorrect order number received was: ' . $callback_data->order_id;

                $mail_body = str_replace('%orderid%', $order_id, $mail_body);
                $mail_body = str_replace('%ordernumber%', $dbOrder->order_number, $mail_body);
                $mail_body = str_replace('%errormsg%', $error_msg, $mail_body);

                $mailer->setSubject('QuickPay Alert - Invalid Order number');
                $mailer->setBody($mail_body);

                $send = $mailer->Send();
                if ( $send !== true ) {
                    echo 'Error sending email: ' . $send->__toString();
                }
            }
            return false;
        }

        // Matches the calculated checksum vs the received checksum
        if(strcasecmp($checksum,$_SERVER['HTTP_QUICKPAY_CHECKSUM_SHA256']) == 0) {
            if($proceed === TRUE) {
                if($test_data == TRUE) {
                    $history->data = '--Test data-- <br />';
                } else {
                    $history->data = NULL;
                }

                // Checks whether the payment was a success
                if($operationData[$operRunID]->qp_status_code == 20000) {
                    $notify = FALSE;

                    // Depending on the callback type, modify order / notify the user/admin
                    switch($operationData[$operRunID]->type) {
                        case "refund":
                            $order_status = $this->payment_params->refunded_status;
                            $notify       = TRUE;
                            break;
                        case "authorize" :
                            $order_status = 'confirmed';
                            $notify       = TRUE;
                            break;
                    }

                    if($notify === TRUE) {
                        $email->subject    = JText::sprintf('PAYMENT_NOTIFICATION_FOR_ORDER','QuickPay',$order_status,$dbOrder->order_number);
                        $email->body       = str_replace('<br/>',"\r\n",JText::sprintf('PAYMENT_NOTIFICATION_STATUS','QuickPay',$order_status)).' '.JText::sprintf('ORDER_STATUS_CHANGED',$order_status)."\r\n\r\n".$order_text;
                        $history->notified = 1;
                    } else {
                        $email             = NULL;
                        $order_status      = NULL;
                        $history->notified = 0;
                    }

                    $history->data .= "QuickPay ID:" . $callback_data->id . "<br />Type: ". ucfirst($operationData[$operRunID]->type) . "<br /> Status: ". $status_codes[$operationData[$operRunID]->qp_status_code];

                    if ($operationData[$operRunID]->type != 'capture' && $notify != false) {
                        $this->modifyOrder($order_id, $order_status, $history, $email);
                    }

                    return true;
                } else {
                    //The order is invalid
                    $order_status       = $this->payment_params->invalid_status;
                    $history->data     .= "QuickPay ID:" . $callback_data->id . "<br />Type: ". ucfirst($operationData[$operRunID]->type) . "<br /> Status: ". $status_codes[$operationData[$operRunID]->qp_status_code];
                    $history->notified  = 1;

                    $email->subject = JText::sprintf('PAYMENT_NOTIFICATION_FOR_ORDER','QuickPay',$order_status,$dbOrder->order_number); //order_id ?
                    $email->body    = str_replace('<br/>',"\r\n",JText::sprintf('PAYMENT_NOTIFICATION_STATUS','QuickPay',$order_status)).' '.JText::sprintf('ORDER_STATUS_CHANGED',$order_status)."\r\n\r\n".$order_text;

                    $this->modifyOrder($order_id, $order_status, $history, $email);

                    return false;
                }
            } else {
                if($alert_on_error == 1){
                    //The order is invalid
                    $order_status       = $this->payment_params->invalid_status;
                    $history->data     .= "QuickPay ID:" . $callback_data->id . "<br />Type: ". ucfirst($operationData[$operRunID]->type) . "<br /> Status: ". $status_codes[$operationData[$operRunID]->qp_status_code];
                    $history->notified  = 1;

                    $email->subject = JText::sprintf('PAYMENT_NOTIFICATION_FOR_ORDER','QuickPay',$order_status,$dbOrder->order_number); //order_id ?
                    $email->body    = str_replace('<br/>',"\r\n",JText::sprintf('PAYMENT_NOTIFICATION_STATUS','QuickPay',$order_status)).' '.JText::sprintf('ORDER_STATUS_CHANGED',$order_status)."\r\n\r\n".$order_text;

                    $this->modifyOrder($order_id, $order_status, $history, $email);

                    //The order is invalid - Test data used
                    $error_msg = 'Test credit card from Quick Pay was attempted used while test_mode was disabled.';
                    $mail_body = str_replace('%orderid%', $order_id, $mail_body);
                    $mail_body = str_replace('%ordernumber%', $dbOrder->order_number, $mail_body);
                    $mail_body = str_replace('%errormsg%', $error_msg, $mail_body);

                    $mailer->setSubject('QuickPay Alert - Test data attempted used');
                    $mailer->setBody($mail_body);
                    $send = $mailer->Send();

                    if ( $send !== true ) {
                        echo 'Error sending email: ' . $send->__toString();
                    }
                } else {
                    if ($operationData[$operRunID]->qp_status_code == 20000) {
                        $notify = FALSE;

                        // Depending on the callback type, modify order / notify the user/admin
                        switch ($operationData[$operRunID]->type) {
                            case "refund":
                                $order_status = $this->payment_params->refunded_status;
                                $notify = TRUE;
                                break;
                            case "authorize" :
                                $order_status = 'confirmed';
                                $notify = TRUE;
                                break;
                        }

                        if ($notify === TRUE) {
                            $email->subject = JText::sprintf('PAYMENT_NOTIFICATION_FOR_ORDER', 'QuickPay', $order_status, $dbOrder->order_number);
                            $email->body = str_replace('<br/>', "\r\n", JText::sprintf('PAYMENT_NOTIFICATION_STATUS', 'QuickPay', $order_status)) . ' ' . JText::sprintf('ORDER_STATUS_CHANGED', $order_status) . "\r\n\r\n" . $order_text;
                            $history->notified = 1;
                        } else {
                            $email = NULL;
                            $order_status = NULL;
                            $history->notified = 0;
                        }

                        $history->data .= "QuickPay ID:" . $callback_data->id . "<br />Type: " . ucfirst($operationData[$operRunID]->type) . "<br /> Status: " . $status_codes[$operationData[$operRunID]->qp_status_code];

                        if ($operationData[$operRunID]->type != 'capture' && $notify != false) {
                            $this->modifyOrder($order_id, $order_status, $history, $email);
                        }

                        return true;
                    }
                }

                return false;
            }
        } else {
            //Checksum failed, if alert_on_error is turned on, send alert email.
            if($alert_on_error == 1){
                $error_msg = 'The checksum calculated did not match the calculated checksum received from QuickPay';

                $mail_body = str_replace('%orderid%', $order_id, $mail_body);
                $mail_body = str_replace('%ordernumber%', $dbOrder->order_number, $mail_body);
                $mail_body = str_replace('%errormsg%', $error_msg, $mail_body);

                $mailer->setSubject('QuickPay Alert - Invalid checksum');
                $mailer->setBody($mail_body);

                $send = $mailer->Send();
                if ( $send !== true ) {
                    echo 'Error sending email: ' . $send->__toString();
                }
            }
            return false;
        }
    }

    function onAfterOrderUpdate(&$order,&$send_email) {
        // Get configuration of HikaShop, and the used payment method
        $config =& hikashop_config();
        if(isset($order->old)) {
            $this->pluginParams($order->old->order_payment_id);
        } else {
            $this->pluginParams();
        }

        $this->payment_params = $this->plugin_params;
        $method_id = $order->old->order_payment_id;

        // Get the status settings
        $cancelled = $this->payment_params->invalid_status;
        $pending   = $this->payment_params->pending_status;
        $confirmed = $this->payment_params->verified_status;
        $refunded  = $this->payment_params->refunded_status;

        // Check what the order status was changed to
        if(isset($order->order_status)) {
            if($order->order_status == $confirmed) {
                $action = 'capture';
            } else if($order->order_status == $refunded) {
                $action = 'refund';
            } else if($order->order_status == $cancelled) {
                $action = 'cancel';
            }
        }

        if(isset($action)) {
            // Get the order's history data.
            // Unfortunate, this seems to be the only place that the transaction ID is logged in the database.
            $db = Factory::getDbo();
            $q = $db->getQuery(true)
                ->select($db->qn('history_data'))
                ->from($db->qn('#__hikashop_history'))
                ->where($db->qn('history_order_id')   . ' = '    . $db->q($order->order_id))
                ->where($db->qn('history_new_status') . ' = '    . $db->q($pending))
                ->where($db->qn('history_data')       . ' LIKE ' . $db->q('%QuickPay ID:%'))
                ->order('history_id DESC')
                ->setLimit('1');
            $res = $db->setQuery($q)->loadResult();



            if($res !== null) {
                // Attempt to extract the transaction ID
                $matches        = array();
                $t              = preg_match('#QuickPay ID:(.*?)\<br \/>#ms', $res, $matches);
                $transaction_id = isset($matches[1]) ? $matches[1] : '0';

                // Only continue if a transaction ID could be extracted
                if($transaction_id !== 0) {
                    $amount = (int)($order->old->order_full_price * 100);

                    // If we're refunding, calculate new amount if shipping fee is not to be included on refunds,
                    // and the order is not set as created/confirmed (aka it's been shipped)
                    if($action == 'refund' && $this->payment_params->refundshipping == 0 && ($order->old->order_status != $pending && $order->old->order_status != $confirmed)) {
                        $amount = (int)(($order->old->order_full_price - $order->old->order_shipping_price) * 100);
                    }
                    $this->vars = array
                    (
                        'action'         => $action,
                        'api_user_key'   => $this->payment_params->api_user_key,
                        'id'             => $transaction_id,
                        'amount'         => $amount
                    );

                    return $this->showPage('api');
                }
            }
        }
    }

    function sign_send($params, $api_key) {
        //Generate checksum to send with the form
        $flattened_params = $this->flatten_params($params);
        ksort($flattened_params);

        $base = implode(" ", $flattened_params);

        return hash_hmac("sha256", $base, $api_key);
    }

    function flatten_params($obj, $result = array(), $path = array()) {
        if (is_array($obj)) {
            foreach ($obj as $k => $v) {
                $result = array_merge($result, $this->flatten_params($v, $result, array_merge($path, array($k))));
            }
        } else {
            $result[implode("", array_map(function($p) { return "[{$p}]"; }, $path))] = $obj;
        }

        return $result;
    }

    function sign_received($base, $private_key) {
        //Generate checksum to check against received checksum
        return hash_hmac("sha256", $base, $private_key);
    }

    function onPaymentConfiguration(&$element) {
		$app = Factory::getApplication();
		$input = $app->input; 
        $subtask = $input->get('subtask', '');
        if($subtask == 'ips') {
            $ips = null;
            echo implode(',', $this->_getIPList($ips));
            exit;
        }

        parent::onPaymentConfiguration($element);
        $this->address = hikashop_get('type.address');

        if(empty($element->payment_params->email)) {
            $app    = Factory::getApplication();
            $lang   = Factory::getLanguage();
            $locale = strtolower(substr($lang->get('tag'), 0, 2));
            //$app->enqueueMessage(JText::sprintf('ENTER_INFO_REGISTER_IF_NEEDED', 'Quickpay', JText::_('HIKA_EMAIL'), 'Quickpay', 'https://quickpay.net/' . $locale . '/mrb/pal=SXL9FKNKGAEM8'));
        }
    }

    function onPaymentConfigurationSave(&$element) {
        if(!empty($element->payment_params->ips))
            $element->payment_params->ips = explode(',', $element->payment_params->ips);
        return true;
    }

    function getPaymentDefaultValues(&$element) {
        $element->payment_name                    = 'Quickpay';
        $element->payment_description             = 'You can pay by Quickpay using this payment method';
        $element->payment_images                  = 'MasterCard,VISA,Credit_card';
        $element->payment_params->url             = 'https://payment.quickpay.net';
        $element->payment_params->cardtypelock    = 'creditcard';
        $element->payment_params->notification    = 1;
        $element->payment_params->ips             = '';
        $element->payment_params->details         = 0;
        $element->payment_params->invalid_status  = 'cancelled';
        $element->payment_params->pending_status  = 'created';
        $element->payment_params->verified_status = 'confirmed';
        $element->payment_params->refunded_status = 'refunded';
        $element->payment_params->autocapture     = 0;
        $element->payment_params->refundshipping  = 0;
        $element->payment_params->splitpayment    = 1;
        $element->payment_params->language        = 'en';
    }

    function _getIPList(&$ipList) {
        $hosts = array(
            'secure.quickpay.dk/form/',
            'secure.quickpay.dk/form/',
            'secure.quickpay.dk/form/'
        );

        $ipList = array();
        foreach($hosts as $host) {
            $ips = gethostbynamel($host);
            if(!empty($ips)) {
                if(empty($ipList))
                    $ipList = $ips;
                else
                    $ipList = array_merge($ipList, $ips);
            }
        }

        if(empty($ipList))
            return $ipList;

        $newList = array();
        foreach($ipList as $k => $ip) {
            $ipParts = explode('.', $ip);
            if(count($ipParts) == 4) {
                array_pop($ipParts);
                $ip = implode('.', $ipParts) . '.*';
            }
            if(!in_array($ip, $newList))
                $newList[] = $ip;
        }
        return $newList;
    }
}